@@ -56,7 +56,35 @@ PyDoc_STRVAR(module_doc,
   bit2_to_bytes(TYPENAME ## _AS_STRING(bit2_a), n_bit2,			\
 		TYPENAME ## _AS_STRING(data_bytes), n);			\
   return data_bytes;							\
+
   
+#define LIBDNA_COMPLFROMBUFFER(TYPENAME)				\
+  Py_ssize_t n_bit2 = TYPENAME ## _GET_SIZE(bit2_a);			\
+  char *string = NULL; /* dummy */					\
+  PyObject *bit2_ca = TYPENAME ## _FromStringAndSize(string, n_bit2);	\
+  bit2_to_bit2compl(TYPENAME ## _AS_STRING(bit2_a), n_bit2,		\
+		    TYPENAME ## _AS_STRING(bit2_ca));			\
+  return bit2_ca;							\
+  
+#define LIBDNA_REVFROMBUFFER(TYPENAME)					\
+  Py_ssize_t n_bit2 = TYPENAME ## _GET_SIZE(bit2_a);			\
+  char *string = NULL; /* dummy */					\
+  PyObject *bit2_ra = TYPENAME ## _FromStringAndSize(string, n_bit2);	\
+  bit2_to_bit2rev(TYPENAME ## _AS_STRING(bit2_a), n_bit2,		\
+		  TYPENAME ## _AS_STRING(bit2_ra));			\
+  return bit2_ra;							\
+
+#define LIBDNA_REVCOMPLFROMBUFFER(TYPENAME)				\
+  Py_ssize_t n_bit2 = TYPENAME ## _GET_SIZE(bit2_a);			\
+  char *string = NULL; /* dummy */					\
+  PyObject *bit2_rca = TYPENAME ## _FromStringAndSize(string, n_bit2);	\
+  bit2_to_bit2compl(TYPENAME ## _AS_STRING(bit2_a), n_bit2,		\
+		    TYPENAME ## _AS_STRING(bit2_rca));			\
+  bit2_to_bit2rev(TYPENAME ## _AS_STRING(bit2_rca), n_bit2,		\
+		  TYPENAME ## _AS_STRING(bit2_rca));			\  
+  return bit2_rca;							\
+
+
 
 static inline Py_ssize_t nbytes_for_bit2(Py_ssize_t n) {
   Py_ssize_t nbytes;
@@ -194,6 +222,136 @@ static PyObject* byte_frombit2bytes(PyObject *self, PyObject *bit2_a)
 }
 
 
+static inline void bit2_to_bit2compl(const char *bit2_ptr, 
+				     const Py_ssize_t bit2_n,
+				     char *bit2compl_prt)
+{
+  Py_ssize_t dna_i, bit2_i;
+  unsigned char cur_byte;
+
+  #ifdef LIBDNA_DEBUG
+  printf("bit2_n: %zd\n", bit2_n);
+  #endif
+
+  dna_i = 0;
+  for (bit2_i=0; bit2_i < bit2_n; bit2_i++)
+    {
+      cur_byte = bit2_ptr[bit2_i];
+      bit2compl_prt[bit2_i] = ~ cur_byte;
+    }
+}
+
+
+PyDoc_STRVAR(bit2complement_frombit2bytes_doc,
+"Takes a bytes object with 2-bit packed DNA values and returns bytes with the DNA complement");
+static PyObject* bit2complement_frombit2bytes(PyObject *self, PyObject *bit2_a)
+{
+  if (PyBytes_Check(bit2_a)) {
+    LIBDNA_COMPLFROMBUFFER(PyBytes);
+  } else {
+    PyErr_Format(PyExc_TypeError, "Parameter must be bytes.");
+    return NULL;
+  }
+}
+PyDoc_STRVAR(bit2complement_frombit2bytearray_doc,
+"Takes a bytearray object with 2-bit packed DNA values and returns the DNA complement still packed.");
+static PyObject* bit2complement_frombit2bytearray(PyObject *self, 
+						  PyObject *bit2_a)
+{
+  if (PyByteArray_Check(bit2_a)) {
+    LIBDNA_COMPLFROMBUFFER(PyByteArray);
+  } else {
+    PyErr_Format(PyExc_TypeError, "Parameter must be bytearray.");
+    return NULL;
+  }
+}
+
+static inline void bit2_to_bit2rev(const char *bit2_ptr, 
+				   const Py_ssize_t bit2_n,
+				   char *bit2rev_prt)
+{
+  Py_ssize_t dna_i, bit2_i;
+  unsigned char lcur_byte, rcur_byte;
+  int bitpos;
+
+#ifdef LIBDNA_DEBUG
+  printf("bit2_n: %zd\n", bit2_n);
+#endif
+
+  const Py_ssize_t half = bit2_n / 2;
+
+  for (bit2_i=0; bit2_i < half; bit2_i++)
+    {
+      lcur_byte = 0;//bit2_ptr[bit2_i];
+      rcur_byte = 0;//bit2_ptr[bit2_n - bit2_i - 1];
+      for (bitpos = 6; bitpos >= 0; bitpos -= 2) {
+	lcur_byte |= ((bit2_ptr[bit2_i] >> (bitpos)) & 0x3) << (6-bitpos);
+	rcur_byte |= ((bit2_ptr[bit2_n - bit2_i - 1] >> (bitpos)) & 0x3) << (6-bitpos);
+      }
+      bit2rev_prt[bit2_n - bit2_i - 1] = lcur_byte;
+      bit2rev_prt[bit2_i] = rcur_byte;
+    }
+
+  if (bit2_n & 0x1 == 1) {
+    lcur_byte = 0;
+    for (bitpos = 6; bitpos >= 0; bitpos -= 2) {
+      lcur_byte |= ((bit2_ptr[half] >> (bitpos)) & 0x3) << (6-bitpos);
+    }
+    bit2rev_prt[half] = lcur_byte;
+  }
+}
+PyDoc_STRVAR(bit2rev_frombit2bytes_doc,
+"Takes a bytes object with 2-bit packed DNA values and returns the reverse sequence.");
+static PyObject* bit2rev_frombit2bytes(PyObject *self, 
+				       PyObject *bit2_a)
+{
+  if (PyBytes_Check(bit2_a)) {
+    LIBDNA_REVFROMBUFFER(PyBytes);
+  } else {
+    PyErr_Format(PyExc_TypeError, "Parameter must be bytes.");
+    return NULL;
+  }
+}
+PyDoc_STRVAR(bit2rev_frombit2bytearray_doc,
+"Takes a bytearray object with 2-bit packed DNA values and returns the reverse sequence.");
+static PyObject* bit2rev_frombit2bytearray(PyObject *self, 
+					   PyObject *bit2_a)
+{
+  if (PyByteArray_Check(bit2_a)) {
+    LIBDNA_REVFROMBUFFER(PyByteArray);
+  } else {
+    PyErr_Format(PyExc_TypeError, "Parameter must be bytearray.");
+    return NULL;
+  }
+}
+
+PyDoc_STRVAR(bit2revcompl_frombit2bytes_doc,
+"Takes a bytes object with 2-bit packed DNA values and returns the reverse complement sequence still packed.");
+static PyObject* bit2revcompl_frombit2bytes(PyObject *self, 
+					    PyObject *bit2_a)
+{
+  if (PyBytes_Check(bit2_a)) {
+    LIBDNA_REVCOMPLFROMBUFFER(PyBytes);
+  } else {
+    PyErr_Format(PyExc_TypeError, "Parameter must be bytes.");
+    return NULL;
+  }
+}
+PyDoc_STRVAR(bit2revcompl_frombit2bytearray_doc,
+"Takes a bytearray object with 2-bit packed DNA values and returns the reverse complement sequence still packed.");
+static PyObject* bit2revcompl_frombit2bytearray(PyObject *self, 
+						PyObject *bit2_a)
+{
+  if (PyByteArray_Check(bit2_a)) {
+    LIBDNA_REVCOMPLFROMBUFFER(PyByteArray);
+  } else {
+    PyErr_Format(PyExc_TypeError, "Parameter must be bytearray.");
+    return NULL;
+  }
+}
+
+
+
 
 PyDoc_STRVAR(slice_frombit2bytes_doc,
 "Takes a bytes object with 2-bit packed DNA values, a slice, and returns the DNA segment in the slice as bytes. As with other Python slices the last element is not included.");
@@ -408,20 +566,32 @@ static PyObject* slice_frombit2bytearray(PyObject *self, PyObject *args)
 
 
 static PyMethodDef libxsq_methods[] = {
-  {"bit2_frombytes",   (PyCFunction)bit2_frombytes,  METH_O,
+  {"bit2_frombytes", (PyCFunction)bit2_frombytes,  METH_O,
    bit2_frombytes_doc},
-  {"bit2_frombytearray",   (PyCFunction)bit2_frombytearray,  METH_O,
+  {"bit2_frombytearray", (PyCFunction)bit2_frombytearray,  METH_O,
    bit2_frombytearray_doc},
-  {"bit2_frombuffer",   (PyCFunction)bit2_frombuffer,  METH_O,
+  {"bit2_frombuffer", (PyCFunction)bit2_frombuffer,  METH_O,
    bit2_frombuffer_doc},
-  {"byte_frombit2bytes",   (PyCFunction)byte_frombit2bytes,  METH_O,
+  {"byte_frombit2bytes", (PyCFunction)byte_frombit2bytes,  METH_O,
    byte_frombit2bytes_doc},
-  {"byte_frombit2bytearray",   (PyCFunction)byte_frombit2bytearray,  METH_O,
+  {"byte_frombit2bytearray", (PyCFunction)byte_frombit2bytearray,  METH_O,
    byte_frombit2bytearray_doc},
-  {"slice_frombit2bytes",   (PyCFunction)slice_frombit2bytes,  METH_VARARGS,
+  {"slice_frombit2bytes", (PyCFunction)slice_frombit2bytes,  METH_VARARGS,
    slice_frombit2bytes_doc},
-  {"slice_frombit2bytearray",   (PyCFunction)slice_frombit2bytearray,  METH_VARARGS,
+  {"slice_frombit2bytearray", (PyCFunction)slice_frombit2bytearray,  METH_VARARGS,
    slice_frombit2bytearray_doc},
+  {"bit2complement_frombit2bytes", (PyCFunction)bit2complement_frombit2bytes,  METH_O,
+   bit2complement_frombit2bytes_doc},
+  {"bit2complement_frombit2bytearray", (PyCFunction)bit2complement_frombit2bytearray,  METH_O,
+   bit2complement_frombit2bytearray_doc},
+  {"bit2reverse_frombit2bytes", (PyCFunction)bit2rev_frombit2bytes,  METH_O,
+   bit2rev_frombit2bytes_doc},
+  {"bit2reverse_frombit2bytearray", (PyCFunction)bit2rev_frombit2bytearray,  METH_O,
+   bit2rev_frombit2bytearray_doc},
+  {"bit2reversecomplement_frombit2bytes", (PyCFunction)bit2revcompl_frombit2bytes,  METH_O,
+   bit2revcompl_frombit2bytes_doc},
+  {"bit2reversecomplement_frombit2bytearray", (PyCFunction)bit2revcompl_frombit2bytearray,  METH_O,
+   bit2revcompl_frombit2bytearray_doc},
   {NULL,                NULL}           /* sentinel */
 };
 
