@@ -149,7 +149,7 @@ def htmlreport():
                         dest = 'sample_percent',
                         default = 0.05,
                         type = float,
-                        help = 'Percentage of the data to sample (a number between 0 and 1, default: 0.05).')
+                        help = 'Percentage of the data to sample (a number between 0 and 1, default: 0.05, that is 5%).')
     parser.add_argument('-V', '--version',
                         action = 'version',
                         version = ngs_plumbing.__version__)
